# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False

from ..interface import Contract, ContractNotRespected, add_prefix
from ..pyparsing_utils import myOperatorPrecedence
from ..syntax import ParsingTmp, W, opAssoc, simple_contract
from .extensions import Extension
from .suggester import create_suggester

NOT_GLYPH = "!"
AND_GLYPH = ","
OR_GLYPH = "|"


class Logical(object):
    def __init__(self, glyph, precedence):
        self.glyph = glyph
        self.precedence = precedence

    def __str__(self):
        s = self.glyph.join(self._convert(x) for x in self.clauses)
        return s

    def _convert(self, x):
        if isinstance(x, Logical) and x.precedence < self.precedence:
            return "(%s)" % x
        return "%s" % x


class OR(Logical, Contract):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) >= 2
        Contract.__init__(self, where)
        Logical.__init__(self, OR_GLYPH, 1)
        self.clauses = clauses

    def _check_quick(self, context, value):
        """Returns True if this checks out."""

        # first make a quick pass
        for c in self.clauses:
            try:
                # try with fake context
                c._check_contract(context.copy(), value, silent=True)
                # if ok, do with main context
                c._check_contract(context, value, silent=True)
                return True
            except ContractNotRespected as e:
                pass

        return False

    def check_contract(self, context, value, silent):
        orig = context.copy()

        if self._check_quick(context, value):
            return
        else:
            if silent:
                msg = "(Error description suppressed.)"
                raise ContractNotRespected(
                    contract=self, error=msg, value=value, context=context
                )

            # otherwise need to do it again with detailed error messages
            self.get_error(orig, value)

    def get_error(self, context, value):
        """This assumes that we are going to fail"""
        exceptions = []
        for c in self.clauses:
            try:
                # try with fake context
                c._check_contract(context.copy(), value, silent=False)
                # if ok, do with main context
                c._check_contract(context, value, silent=False)

                assert False, "We should not be here."
            except ContractNotRespected as e:
                exceptions.append((c, e))
        else:
            msg = self._format_exceptions(exceptions)
            raise ContractNotRespected(
                contract=self, error=msg, value=value, context=context
            )

    def _format_exceptions(self, exceptions):
        msg = "Could not satisfy any of the %d clauses in %s." % (
            len(self.clauses),
            self,
        )

        for i, ex in enumerate(exceptions):
            c, e = ex
            if i + 1 > 1:
                msg += "\n ├┄┄┄ Clause #%d:   %s\n" % (i + 1, c)
            else:
                msg += "\n ┬┄┄┄┄ Clause #%d:   %s\n" % (i + 1, c)
            msg += add_prefix("%s" % e, " | ")

        msg += "\n ╰┄┄┄┄┄┄ (end clauses) ┄┄┄┄┄┄"
        return msg

    def __repr__(self):
        s = "OR(%r)" % self.clauses
        return s

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        clauses = [l.pop(0)]
        while l:
            glyph = l.pop(0)  # @UnusedVariable
            assert glyph == OR_GLYPH
            operand = l.pop(0)
            clauses.append(operand)
        where = W(string, location)
        return OR(clauses, where=where)


class And(Logical, Contract):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) >= 2, clauses
        Contract.__init__(self, where)
        Logical.__init__(self, AND_GLYPH, 2)
        self.clauses = clauses

    def check_contract(self, context, value, silent):
        for c in self.clauses:
            c._check_contract(context, value, silent)

    def __repr__(self):
        s = "And(%r)" % self.clauses
        return s

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        clauses = [l.pop(0)]
        while l:
            glyph = l.pop(0)  # @UnusedVariable
            assert glyph == AND_GLYPH
            operand = l.pop(0)
            clauses.append(operand)
        where = W(string, location)
        return And(clauses, where=where)


class Not(Logical, Contract):
    def __init__(self, clauses, where=None):
        assert isinstance(clauses, list)
        assert len(clauses) == 1, clauses
        Contract.__init__(self, where)
        Logical.__init__(self, NOT_GLYPH, 3)
        self.clauses = clauses

    def check_contract(self, context, value, silent):
        clause = self.clauses[0]
        try:
            clause._check_contract(context, value, silent)
        except ContractNotRespected:
            pass
        else:
            msg = "Shouldn't have satisfied the clause %s." % clause
            raise ContractNotRespected(
                contract=self, error=msg, value=value, context=context
            )

    @staticmethod
    def parse_action(string, location, tokens):
        l = list(tokens[0])
        assert l.pop(0) == NOT_GLYPH
        where = W(string, location)
        return Not(l, where=where)

    def __repr__(self):
        s = "Not(%r)" % self.clauses
        return s

    def __str__(self):
        return self.glyph + self._convert(self.clauses[0])


suggester = create_suggester(
    get_options=lambda: ParsingTmp.keywords + list(Extension.registrar.keys())
)
baseExpr = simple_contract | suggester
baseExpr.setName("Simple contract (recovering)")


op = myOperatorPrecedence
# op = operatorPrecedence
composite_contract = op(
    baseExpr,
    [
        (NOT_GLYPH, 1, opAssoc.RIGHT, Not.parse_action),
        (AND_GLYPH, 2, opAssoc.LEFT, And.parse_action),
        (OR_GLYPH, 2, opAssoc.LEFT, OR.parse_action),
    ],
)
composite_contract.setName("NOT/OR/AND contract")

or_contract = op(
    baseExpr,
    [
        (OR_GLYPH, 2, opAssoc.LEFT, OR.parse_action),
    ],
)
or_contract.setName("OR contract")



import unittest

from Aspidites._vendor.contracts.docstring_parsing import (
    DocStringInfo,
    Arg,
    number_of_spaces,
)
from Aspidites._vendor.contracts.interface import add_prefix


examples = {
    """
        Provides a RGB representation of the values by interpolating the range
        [min(value),max(value)] into the colorspace [min_color, max_color].

        :param value: The field to represent.
        :type  value: HxW array

        :param max_value: If specified, everything *above* is clipped.
        :type max_value: float
        :param min_value: If specified, everything *below* is clipped.
        :type min_value: float

        :param min_color:  Color to give to the minimum values.


        :return: A RGB image.
        :rtype: HxWx3 uint8

        :return: gray

    """: DocStringInfo(
        docstring="\n        Provides a RGB representation of the values by interpolating the range\n"
        "        [min(value),max(value)] into the colorspace [min_color, max_color].\n",
        params={
            "value": Arg("The field to represent.", "HxW array"),
            "max_value": Arg("If specified, everything *above* is clipped.", "float"),
            "min_value": Arg("If specified, everything *below* is clipped.", "float"),
            "min_color": Arg("Color to give to the minimum values.", None),
        },
        returns=[Arg("A RGB image.", "HxWx3 uint8"), Arg("gray", None)],
    )
}


class DocStringTest(unittest.TestCase):
    def test_parsing(self):
        for string in examples:
            parsed = DocStringInfo.parse(string)
            "%s" % parsed
            "%r" % parsed
            result = examples[string]
            self.assertEqual(result, parsed)

    def test_number_of_spaces(self):
        self.assertEqual(number_of_spaces(""), 0)
        self.assertEqual(number_of_spaces(" "), 1)
        self.assertEqual(number_of_spaces("  "), 2)
        self.assertEqual(number_of_spaces("11"), 0)
        self.assertEqual(number_of_spaces(" 223"), 1)
        self.assertEqual(number_of_spaces("  4343"), 2)

    def test_reparsing(self):
        for string, result in examples.items():  # @UnusedVariable
            parsed = DocStringInfo.parse(string)
            converted = "%s" % parsed
            reparsed = DocStringInfo.parse(converted)

            msg = "First string:\n%s\nParsed as:\n%s\n" % (
                add_prefix(string, "|"),
                add_prefix("%r" % parsed, "|"),
            )

            msg += "Converted:\n%s\nReparsed as:\n%s\n" % (
                add_prefix(converted, "|"),
                add_prefix("%r" % reparsed, "|"),
            )

            self.assertEqual(parsed, reparsed, msg=msg)

    def test_inline_params(self):
        def test_inline_parsing(docstring, expected_type="type", expected_desc="desc"):
            info = DocStringInfo.parse(docstring)
            self.assertTrue("name" in info.params)
            self.assertEqual(info.params["name"].type, expected_type)
            self.assertEqual(info.params["name"].desc, expected_desc)

        # Proper syntax
        test_inline_parsing(":param type name: desc")
        test_inline_parsing(":param name: desc", None)
        test_inline_parsing(":param name:", None, None)

        # Weird syntax for people who like to break things.
        test_inline_parsing(" : param type name : desc ")
        test_inline_parsing(" : param  name : desc ", None)
        test_inline_parsing(" : param  name : ", None, None)
        test_inline_parsing(" : param type , > 0  name : ", "type , > 0", None)

    def test_inline_returns(self):
        def test_inline_parsing(docstring, expected_type="type", expected_desc="desc"):
            info = DocStringInfo.parse(docstring)
            self.assertTrue(len(info.returns) > 0)
            self.assertEqual(info.returns[0].type, expected_type)
            self.assertEqual(info.returns[0].desc, expected_desc)

        # Proper syntax
        test_inline_parsing(":returns type: desc")
        test_inline_parsing(":returns: desc", None)
        test_inline_parsing(":returns:", None, None)

        # Weird syntax for people who like to break things.
        test_inline_parsing(" : returns type : desc ")
        test_inline_parsing(" : returns : desc ", None)
        test_inline_parsing(" : returns : ", None, None)
        test_inline_parsing(" : returns type , > 0 : ", "type , > 0", None)



""" Other testing examples """

from Aspidites._vendor.contracts.test_registrar import good, fail

# dummy
good(["*"], [0])
good(["*", "*"], [0, 1])
good(["=0", "=1"], [0, 1])
fail(["=0", "=1"], [0, 2])

# Lists of equal length
good(["list[N]", "list[N]"], [[4], [3]])
good(["list[N]", "list[N]"], [[], []])
fail(["list[N]", "list[N]"], [[], [1]])
good(["list[N]", "list[N],N>0"], [[1], [3]])
# we can also refer to the other context
good(["list[N]", "list,N>0"], [[1], [3]])
fail(["list[N]", "list,N>0"], [[], [3]])

# Lists of different length
good(["list[N]", "list[M],M!=N"], [[4], [3, 2]])
good(["list[N]", "list[M],M!=N"], [[4, 3], [3]])
fail(["list[N]", "list[M],M!=N"], [[3], [3]])

# One list shorter than the other
good(["list[N]", "list[M],M<N"], [[4, 3], [3]])
good(["list[N]", "list[M],N>M"], [[4, 3], [3]])
fail(["list[N]", "list[M],N>M"], [[3], [3]])

# Values of the same type
good(["type(x)", "type(x)"], [0, 1])
good(["type(x)", "type(x)"], [0.1, 1.1])
fail(["type(x)", "type(x)"], [0.1, 1])
good(["type(x)", "type(y)", "type(x)|type(y)"], [0, 1.1, 1.2])
good(["type(x)", "type(y)", "type(x)|type(y)"], [0, 1.1, 3])
fail(["type(x)", "type(y)", "type(x)|type(y)"], [0, 1.1, "ciao"])

# A list with all elements of the same type
good(["list(type(x))"], [[0, 1, 2]])
fail(["list(type(x))"], [[0, 1.2, 2]])
fail(["list(type(x))"], [[0, None, 2]])
# Lists with the same type of elements
good(["list(type(x))", "list(type(x))"], [[1, 2], [3, 4]])
fail(["list(type(x))", "list(type(x))"], [[1, 2], [3.0, 4]])
# Using different variables and then imposing they are equal
good(["list(type(x))", "list(type(y)),x=y"], [[1, 2], [3, 4]])
fail(["list(type(x))", "list(type(y)),x=y"], [[1, 2], [3.0, 4]])

# a list with at most two types
good("list(type(x|y))", [1, 2, 3, 4.0])
fail("list(type(x|y))", [1, 2, "ciao", 4.0])
